#!/usr/bin/python3

import unittest
import doctest
import ucammap

# A 'usual' test address
ADDRESS = [
    {'type': 'location', 'location': 987},
    {'type': 'site',     'site': 'New Museums Site'}, 
    {'type': 'number',   'number': '123'},
    {'type': 'number',   'number': '456'},
    {'type': 'street',   'street': 'Pembroke Street'}, 
    {'type': 'place',    'place': 'Cambridge'}, 
    {'type': 'postcode', 'postcode': 'CB2 3QH'},
    {'type': 'number',   'number': '999'},
    {'type': 'address',  'address': 'Some unparsed text'}
    ]

# An address without structured data
ADDRESS2 = [
    {'type': 'location', 'location': 987},
    {'type': 'address',  'address': 'Some unparsed text'}
    ]

# An address with no useful data
ADDRESS3 = [
    {'type': 'location', 'location': 987},
    ]

CONTACT = {'type': 'url',
           'url': 'http://www.ucs.cam.ac.uk/about-us/contact/general', 
           'role': 'contacts', 'options': 'foo'}

# An info block containing ADDRESS and CONTACT
INFO =  [
    {'type': 'address', 'address': ADDRESS}, 
    {'type': 'phone',   'phone': '+44 1223 334600'}, 
    {'type': 'fax',     'fax': '+44 1223 334679'}, 
    {'type': 'email',   'email': 'reception@ucs.cam.ac.uk'}, 
    {'type': 'url',     
     'url': 'http://www.ucs.cam.ac.uk/', 
     'role': 'webmaster'}, 
    CONTACT,
    ]

# Entity data containing INFO
EXAMPLE = {
    'type': 'institution', 
    'inst': 'cs', 
    'role': 'nonacademic', 
    'prefix': 'University', 
    'name': 'Computing Service', 
    'suffix': 'UCS',
    'parent': {
        'type': 'institution', 
        'inst': 'misd', 
        'name': 'MISD', 
    },
    'locations': [
        {'name': 'Cockcroft Building', 
         'type': 'building', 
         'lon': 0.119814, 
         'lat': 52.203322, 
         'ref': 'M039', 
         'id': 95412331
        }
    ], 
    'entrances': [
        {'direction': 34, 
         'door': True, 
         'entrance': 'main', 
         'type': 'entrance', 
         'lon': 0.119701, 
         'primitive': 'n', 
         'user': 'f', 
         'lat': 52.203379, 
         'ref': 'M038-A', 
         'id': 1443478065
        }, 
        {'direction': -77, 
         'door': False, 
         'entrance': 'main', 
         'type': 'entrance', 
         'lon': 0.120499, 
         'primitive': 'n', 
         'user': 'fbv', 
         'lat': 52.202871, 
         'ref': 'M-A', 
         'id': 1091639792
        }, 
        {'direction': -145, 
         'door': False, 
         'entrance': 
         'secondary', 
         'type': 'entrance', 
         'lon': 0.11877, 
         'primitive': 'n', 
         'user': 'f', 
         'lat': 52.203481, 
         'ref': 'M-C', 
         'id': 1091640188
        }
    ], 
    'sd': 'ucs',
    'info': INFO,
    'debug1': False, 
}

class EntityTestCase(unittest.TestCase):

    def setUp(self):
        self.entity   = ucammap.Entity(EXAMPLE)

    def testEntityRawData(self):
        """Extracting entity raw data"""
        self.assertEqual(self.entity.raw_data,EXAMPLE)
    
    def testEntitySimpleAttributes(self):
        """Retrieving simple entity attributes"""
        self.assertEqual(self.entity.type,'institution')
        self.assertEqual(self.entity.inst,'cs')
        self.assertEqual(self.entity.role,'nonacademic')
        self.assertEqual(self.entity.prefix,'University')
        self.assertEqual(self.entity.name,'Computing Service')
        self.assertEqual(self.entity.suffix,'UCS')
        self.assertEqual(self.entity.sd,'ucs')

    def testEntityMissingAttributes(self):
        """Retreving missing/non-existent atentity attributes"""
        # Missing attributes return None
        self.assertIsNone(self.entity.zoom)
        self.assertIsNone(self.entity.options)
        self.assertIsNone(self.entity.sortvia)
        # ...as do non-existent ones
        self.assertIsNone(self.entity.nonsuch)

    def testEntityHarderAttributes(self):
        """Retrieving list or object entity attributes"""

        # institution/parent are None or ucammap.Entity
        self.assertIsNone(self.entity.institution)

        self.assertIs   (type(self.entity.parent),ucammap.Entity)
        self.assertEqual(self.entity.parent.name,'MISD')

        # locations/entrances/occupants are lists of ucammap.Entity
        self.assertIs   (type(self.entity.locations),list)
        self.assertEqual(len(self.entity.locations),1)
        self.assertIs   (type(self.entity.locations[0]),ucammap.Entity)
        self.assertEqual(self.entity.locations[0].name,'Cockcroft Building')

        self.assertIs   (type(self.entity.entrances),list)
        self.assertEqual(len(self.entity.entrances),3)
        self.assertIs   (type(self.entity.entrances[0]),ucammap.Entity)
        self.assertEqual(self.entity.entrances[0].id,1443478065)

        self.assertIs   (type(self.entity.occupants),list)
        self.assertEqual(len(self.entity.occupants),0)

        # info is a list of ucammap.info
        self.assertIs   (type(self.entity.info),list)
        self.assertEqual(len(self.entity.info),6)
        self.assertIs   (type(self.entity.info[1]),ucammap.Info)
        self.assertEqual(self.entity.info[1].value,'+44 1223 334600')

    def testEntityMethods(self):
        """Entity utility methods"""

        # full_name
        self.assertEqual (self.entity.full_name(),
                           'University Computing Service (UCS)')
        # sort_name
        self.assertEqual (self.entity.sort_name(),
                           'Computing Service, University (UCS)')
        # filtered_info
        info = self.entity.filtered_info('url')
        self.assertIs   (type(info),list)
        self.assertEqual(len(info),2)
        self.assertIs   (type(info[1]),ucammap.Info)

        # __str__
        self.assertEqual (str(self.entity), 
                          'University Computing Service (UCS)')
        
    def testInfo(self):
        """Geneal info object tests"""

        info_data = {
            'type': 'url',
             'url': 'http://www.ucs.cam.ac.uk/about-us/contact/general', 
             'role': 'contacts', 'options': 'foo'}
        info = ucammap.Info(info_data)

        # Info object attributes match test data
        self.assertEqual (info.raw_data, info_data)
        self.assertEqual (info.type,'url')
        self.assertEqual (info.value,
                           'http://www.ucs.cam.ac.uk/about-us/contact/general') 
        self.assertEqual (info.role,'contacts')
        self.assertEqual (info.options,'foo')
        self.assertEqual (str(info),
           'Url: http://www.ucs.cam.ac.uk/about-us/contact/general (contacts)')
        # Non-existent Info attributes shoud be None
        self.assertIsNone (info.nonsuch)

        info_data = {
            'type': 'address', 
             'address':  [
                 {'type': 'location', 'location': 987},
                 {'type': 'site',     'site': 'New Museums Site'}, 
                 {'type': 'number',   'number': '123'},
                 {'type': 'number',   'number': '456'},
                 {'type': 'street',   'street': 'Pembroke Street'}, 
                 {'type': 'place',    'place': 'Cambridge'}, 
                 {'type': 'postcode', 'postcode': 'CB2 3QH'},
                 {'type': 'number',   'number': '999'},
                 {'type': 'address',  'address': 'Some unparsed text'} ] }
        info = ucammap.Info(info_data)

        self.assertEqual (info.raw_data, info_data)
        self.assertEqual (info.type,'address')
        self.assertEqual (type(info.value), ucammap.Address)

        # A info with no type throws an error
        with self.assertRaisesRegexp(ucammap.APIError,
                                             'Type field missing') as cm:
                c = ucammap.Info({'typeZZZ': 'url',
                   'url': 'http://www.ucs.cam.ac.uk/about-us/contact/general', 
                   'role': 'contacts', 'options': 'foo'})

        # A info record without the right data key  throws an error
        with self.assertRaisesRegexp(ucammap.APIError,
                                             'Data field "url" missing') as cm:
            c = ucammap.Info({'type': 'url',
                 'urlZZZ': 'http://www.ucs.cam.ac.uk/about-us/contact/general', 
                 'role': 'contacts', 'options': 'foo'})

    def testAddress(self):
        """General address object tests"""

        address_data = [
            {'type': 'location', 'location': 987},
            {'type': 'site',     'site': 'New Museums Site'}, 
            {'type': 'number',   'number': '123'},
            {'type': 'number',   'number': '456'},
            {'type': 'street',   'street': 'Pembroke Street'}, 
            {'type': 'place',    'place': 'Cambridge'}, 
            {'type': 'postcode', 'postcode': 'CB2 3QH'},
            {'type': 'number',   'number': '999'},
            {'type': 'address',  'address': 'Some unparsed text'}
            ]

        address  = ucammap.Address(address_data)
        self.assertEqual (address.raw_data, address_data)
        self.assertIs    (type(address.as_block()),list)
        self.assertEqual (len(address.as_block()),5)
        self.assertEqual (address.as_line(),
           'New Museums Site, 123 456 Pembroke Street, Cambridge, CB2 3QH, 999')
        self.assertEqual (str(address),
           'New Museums Site, 123 456 Pembroke Street, Cambridge, CB2 3QH, 999')

        # An address without structured data
        address2_data = [
            {'type': 'location', 'location': 987},
            {'type': 'address',  'address': 'Some unparsed text'}
            ]

        address2 = ucammap.Address(address2_data)
        self.assertEqual (address2.raw_data, address2_data)
        self.assertIs    (type(address2.as_block()),list)
        self.assertEqual (len(address2.as_block()),0)
        self.assertEqual (address2.as_line(),'Some unparsed text')

        # An address with no useful data
        address3_data = [
            {'type': 'location', 'location': 987},
            ]

        address3 = ucammap.Address(address3_data)
        self.assertEqual (address3.raw_data, address3_data)
        self.assertIs    (type(address3.as_block()),list)
        self.assertEqual (len(address3.as_block()),0)
        self.assertIsNone(address3.as_line())

        # A address containing an entry with no type throws an error
        with self.assertRaisesRegexp(ucammap.APIError,
                                     'Type field missing') as cm:
            c = ucammap.Address([{'typeZZZ': 'location', 'location': 987}])

        # A address containing an entry with no data field throws an error
        with self.assertRaisesRegexp(ucammap.APIError,
                                     'Data field "location" missing') as cm:
            c = ucammap.Address([{'type': 'location', 'locationZZZ': 987}])

    def testMinimal(self):
      """Correct handling of other missing entity data"""

      minimal_data = {
          'type': 'institution', 
          'name': 'Computing Service', 
      }

      entity = ucammap.Entity(minimal_data)
      self.assertEqual(entity.raw_data,minimal_data)

      # info should be returning an empty list
      self.assertIs   (type(entity.info),list)
      self.assertEqual(len(entity.info),0)

      # filtered_info should also return an empty list
      self.assertIs   (type(entity.filtered_info('url')),list)
      self.assertEqual(len(entity.filtered_info('url')),0)


if __name__ == '__main__':
    unittest.main()
