"""
A pythonic interface to the University of Cambridge Map database API.

This is a thin-ish wrapper around the map database API - see
https://wiki.cam.ac.uk/university-map/The_Database_API for more
details of the API.

This module implements the following classes:
API       -- an API for querying map database
Entity    -- a representation of the entities (institutions, buildings, etc.)
             returned by the api
Info      -- a representation of contact information returned
Address   -- a representation of address information returned
APIError  -- a local exception object for the class 

Note the following design features:
  * All objects produce something useful when converted to a string
    (and so when printed), which is useful for experimentation
  * Objects contain methods that return their content in common formats
  * Attributes listed in the API documentation are available as 
    attributes of the corresponding objects
  * Attributes evaluate as False if not available, making this 
    pattern easy -
       if object.thing:
           print(<header for thing>)
           <do stuff with thing>
    Note that this happens for both 'attribute not available in this type'
    and 'this instance doesn't have this attribute'  
  * The raw data that initialized any particular object is available
    as its raw_data attribute

Basic usage:
>>> import ucammap
>>> api = ucammap.API()
>>> for result in api.search(q = 'computing'):
...     print(result.full_name())
... 
Department of Applied Mathematics and Theoretical Physics (DAMTP)
Bioinformatics and Computational Biology Services
Cambridge Computational Biology Institute (CCBI)
Department of Chemistry
Collaborative Computing Project for NMR
Computer Laboratory
University Computing Service
High Performance Computing Service (HPCS)

>>> import ucammap
>>> api = ucammap.API()
>>> for entity in api.search(inst = 'cs'):
...   for info in entity.filtered_info('address'):
...     print(info.value.as_line())
... 
New Museums Site, Pembroke Street, Cambridge, CB2 3QH

"""

import urllib.request
import urllib.parse
import json

import logging

_logger = logging.getLogger(__name__)


class API:

    """
    An interface to the Map API.

    Provides basic searching and retrieval functions for retrieving
    information from the database API.

    Implements:
       __init__ 
       search   -- Search for entities

    """

    def __init__(self, base='http://map.cam.ac.uk/v5.json'):
        """
        Construct interface.

        Keyword arguments:
        base -- URL of the interface (default 'http://map.cam.ac.uk/v5.json')

        """
        self.base = base

    def search(self, **kwargs):
        """
        Search database, return matching entities.

        Keyword arguments: all keyword arguments are converted to URL
        parameters and included in a call to the underlying interface.

        Returns a (possibly empty) sequence of Entity objects.

        Raises APIError following error return from the
        interface. May propagates exceptions from urllib if the
        underlying HTTP request fails.

        """
        
        parameters  = urllib.parse.urlencode(kwargs)
        _logger.info("search called with %s",kwargs) 

        response = None
        try:
            url = self.base + '?' + parameters
            _logger.debug("Requesting %s",url)
            response = urllib.request.urlopen(url)
        except urllib.error.HTTPError as e:
            if e.code == 400 and e.info()["Content-type"] == 'application/json':
                # Status = 400 + JSON content is API error - use content
                response = e
            else:
                # ...otherwise propagate
                raise

        _logger.debug("Response code %s, type %s", 
                    response.code,response.info()["Content-type"])   

        # Something badly wrong if response not application/json  
        content_type = response.info()["Content-type"]
        if content_type != 'application/json':
            raise APIError('Unexpected content returned: ' + content_type)

        # Encoding not specified - rely on default utf-8
        data = json.loads(response.read().decode('utf-8'))
        _logger.debug("Response content %s",data)

        results = []
        for item in data:
            if item['type'] == 'error':
                # Convert returned error entities to exceptions
                raise APIError(item['error'])
            else:
                # ...otherwise convert to entity
                results.append(Entity(item))

        _logger.info("Successfully returned %s entities",len(results))        

        return results


class Entity:

    """
    A representation of a database entity (institution, building, etc.).

    Implements:
    __init__
    full_name()
    sort_name()
    __getattr__
    __str__
    
    Instance variables:
    raw_data -- underlying data structure returned by the API as a dictionary
    
    In addition, __getattr__ makes all API data structure elements
    available as attributes (entity.name, entity.logo, etc.).

    """
    def __init__(self,data):
        """Construct new entity object from dictionary representation"""
        self.raw_data = data

    def __getattr__(self,name):
 
        """
        Return any API data structure element as an attribute.

        Simple elements are returned as strings, or None if not
        present; elements relating to single entities ('parent',
        'institution') are returned as an Entity object, or None of
        not present; elements relating to multiple entities
        ('locations', 'entrances', 'occupants') are returned as a
        (possibly empty) sequence of Entity objects; references to
        contact info ('info') are returned as a (possibly empty)
        sequence of Info objects.

        Note that all elements evaluate as false if not available, making
 
          if entity.thing:
            <do stuff with thing> 

        easy. Note also that this doesn't distinguish between elements
        documented in the API and anything else - both return None. 

        """
    
        # Fields that are sequences of entities
        if name in ('locations', 'entrances', 'occupants'):
            entities = []
            if name in self.raw_data:
                for item in self.raw_data[name]:
                    entities.append(Entity(item))
            return entities

        # Fields that are single entities
        elif name in ('parent','institution') and name in self.raw_data: 
            return (Entity(self.raw_data[name]))

        # Info, which is a sequence of info objects
        elif name == 'info':
            info = []
            if 'info' in self.raw_data:
                for item in self.raw_data['info']:
                    info.append(Info(item))
            return info
            
        # Simple fields
        elif name in self.raw_data:
            return self.raw_data[name]

        # Else None
        else:
            return None

    def full_name(self):
        """Return a full human-readable name"""
        value = []
        if self.prefix:
            value.append(self.prefix)
        if self.name:
            value.append(self.name)
        if self.suffix:
            value.append('(' + self.suffix + ')')
        return ' '.join(value)   

    def sort_name(self):
        """Return a human-readable name suitable for sorted lists"""
        value = []
        if self.name:
            value.append(self.name)
        if self.prefix:
            value.append(self.prefix)
        value = [', '.join(value)]
        if self.suffix:
            value.append('(' + self.suffix + ')')
        return ' '.join(value)    

    def filtered_info(self,key):
        """Return only Info records of type matching the supplied key."""
        return [c for c in self.info if c.type == key]

    def info_keys(self):
        """Return a sorted list of unique contect types for the entity."""
        s = set([info.type for info in self.info])
        return sorted(list(s))

    def __str__(self):
        """Return a human-readable representation"""
        return self.full_name()


class Info:
    """A container for a single contact info item.

    Implements:
    __init__
    __getattr__
    __str__

    Instance variables:
    raw_data -- underlying data structure returned by the API as a 
                sequence of dicts

    In addition, __getattr__ makes the item's 'type', 'value', 'role'
    and 'options' available as attributes. Note that info of type
    'address' is returned as an Address object.

    """

    def __init__(self,data):
        """
        Construct new Info object from dictionary representation

        Raises APIError if called with a data structure that doesn't
        contain a 'type' field or which doesn't contain a data field
        named after the type field.

        """

        if not 'type' in data:
            raise APIError('Type field missing from info record ' + str(data))
        info_type = data['type']
        if not info_type in data:
            raise APIError('Data field "' + info_type + 
                           '" missing from info record ' + str(data))
        self.raw_data = data

    def __getattr__(self,name):
        """Return attributes from an Info record"""
        if name == 'value':
            info_type = self.raw_data['type']
            if info_type == 'address':
                return Address(self.raw_data['address'])
            else:
                return self.raw_data[info_type]
        elif name in self.raw_data:
            return self.raw_data[name]
        else:
            return None
        
    def __str__(self):
        """Return a useful string representation of value + role"""
        result = self.type.capitalize() + ': ' + str(self.value)
        if self.role:
            result += ' (' + self.role + ')'
        return result    


class Address:

    """
    A representation of a structured address

    Implements:
    __init__ 
    as_line()   -- address as one line
    as_block()  -- address as a sequence of lines
    __str__

    Instance variables:
    data  -- the underlying API data structure (a sequence of one element dicts)
 
    """

    def __init__(self,address):
        """
        Construct address object from dictionary

        Raises APIError if called with a data structure that doesn't
        contain a 'type' field or which doesn't contain a data field
        named after the type field.

        """

        for line in address:
            if not 'type' in line:
                raise APIError('Type field missing from address record ' 
                               + str(line))
            address_type = line['type']
            if not address_type in line:
                raise APIError('Data field "' + address_type + 
                               '" missing from address record ' + str(line))
            self.raw_data = address

    def lines(self):
        """Return a sequence of (type,value) sequences"""
        result  = []
        for line in self.raw_data:
            address_type = line['type']
            result.append((address_type,line[address_type]))
        return result

    def as_block(self):
        """Provide address as an array of line items"""
        result  = []
        numbers = []
        for (key,value) in self.lines():
            if key in ('location', 'address'):
                pass
            elif key == 'number':
                numbers.append(value)
            else:
                if numbers:
                    result.append(' '.join(numbers) + ' ' + value)
                    numbers = []
                else:
                    result.append(value)
        if numbers:
            result.append(' '.join(numbers))
        return result

    def as_line(self):
        """Provide address as one line"""
        block = self.as_block()
        if block:
            return ", ".join(block)
        # ... otherwise return 'address' if available
        for (key,value) in (self.lines()):
            if key == 'address':
                return value
        # ... otherwise None
        return None

    def __str__(self):
        """Address suitable for simple print"""
        return self.as_line()


class APIError(Exception):
    """Custom error class for the ucammap module"""

    def __init__(self,reason):
        self.reason = reason

    def __str__(self):
        return self.reason


if __name__ == "__main__":
    import doctest
    doctest.testmod()
